import numpy as np
from datetime import timedelta
import matplotlib.dates as mdates


def plot_acc(ax, acc, start_time):
    sf = 50
    win = 15

    diff_acc = np.abs(acc[:, 1:] - acc[:, 0:-1])
    diff_acc = np.c_[diff_acc, [0, 0, 0]]

    avg_diff_acc = np.sum(np.reshape(np.sum(diff_acc, axis=0), [-1, sf * win]), axis=1) / (sf * win)
    # set max diff acc to 500
    avg_diff_acc[avg_diff_acc > 500] = 500
    data_length = avg_diff_acc.shape[0]

    t = np.arange(data_length) * win / 3600
    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)
    ax.plot(timestamp_num, avg_diff_acc, lw=1.5, color='r')

    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_ylabel("Head Movement", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim(timestamp[0], timestamp[-1])


import numpy as np


def load_sti(sti_path):
    index_list = []
    with open(sti_path, 'r') as f:
        for line in f:
            if line.startswith("point count: "):
                index = line.split("\t")[0][13:]
                index_list.append(int(index) // 5)
    index_list = np.array(index_list, dtype=np.int32) if len(index_list) >= 2 else None
    return index_list


if __name__ == '__main__':
    sti_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\sti.log'
    sti_data = load_sti(sti_path)



"""
绘制分期图
"""
import matplotlib.pyplot as plt
from qleegss.plot.plot_eeg_spectrogram import plot_spectrogram
import numpy as np
from datetime import timedelta, datetime
import matplotlib.dates as mdates
from scipy.signal import savgol_filter
from qleegss.plot.plot_acc import plot_acc


def plot_stage(eeg, eeg_start_time, sf_eeg, eeg_path, acc, stage_res):
    fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1, figsize=(20, 4 * 4))
    # spectrogram
    t = plot_spectrogram(ax1, eeg, eeg_start_time, sf_eeg)
    # acc
    plot_acc(ax2, acc, eeg_start_time)
    # posture
    plot_sleep_posture(ax3, sleep_posture_analyse(acc), eeg_start_time)
    # stage
    plot_stage_res(ax4, stage_res, eeg_start_time)

    # config
    plt.tight_layout()
    save_path = eeg_path.replace('eeg.eeg', 'sleep_fig.png')
    plt.savefig(save_path, dpi=300, bbox_inches='tight')


def plot_sleep_posture(ax, grade, start_time):
    sf = 50
    # assert grade.shape[0] == 1, "The grade of head bias should be a 1-D ndarray"
    t = np.arange(grade.shape[0]) / sf / 3600
    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)
    ax.plot(timestamp_num, grade, lw=1.5, color='b')
    ax.set_ylim(-3.5, 3.5)
    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_yticks([-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi])
    ax.set_yticklabels(['Sleep Face Down', 'Lie on the Left', 'Lie Flat', 'Lie on the Right', 'Sleep Face Down'], )
    ax.set_ylabel("Sleep Postures", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})
    ax.grid(visible=True, axis='y', linewidth=0.5)

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim(timestamp[0], timestamp[-1])


def sleep_posture_analyse(acc):
    acc_y = acc[1, :]
    acc_z = acc[2, :]

    cos = acc_z / np.sqrt(acc_z * acc_z + acc_y * acc_y)
    upper_grade = np.arccos(cos)
    grade = upper_grade * (acc_y / (np.abs(acc_y)+1e-16))
    grade = savgol_filter(grade, window_length=50, polyorder=1)

    return grade


def plot_stage_res(ax, hypno, start_time):
    win = 30

    t = np.arange(hypno.size) * win / 3600
    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)

    map_dict = {
        3: 0,
        2: 1,
        1: 2,
        4: 3,
        0: 4,
        5: 5
    }
    #
    hypno = [map_dict.get(x, x) for x in hypno]

    n3_sleep = np.ma.masked_not_equal(hypno, 0)
    n2_sleep = np.ma.masked_not_equal(hypno, 1)
    n1_sleep = np.ma.masked_not_equal(hypno, 2)
    rem_sleep = np.ma.masked_not_equal(hypno, 3)
    wake = np.ma.masked_not_equal(hypno, 4)
    abnormal = np.ma.masked_not_equal(hypno, 5)

    ax.plot(timestamp_num, hypno, lw=2, color='k')
    ax.plot(timestamp_num, abnormal, lw=2, color='k')
    ax.plot(timestamp_num, wake, lw=2, color='orange')
    ax.plot(timestamp_num, rem_sleep, lw=2, color='lime')
    ax.plot(timestamp_num, n1_sleep, lw=2, color='yellowgreen')
    ax.plot(timestamp_num, n2_sleep, lw=2, color='deepskyblue')
    ax.plot(timestamp_num, n3_sleep, lw=2, color='royalblue')

    ax.set_ylim([-0.1, 5.8])
    ax.set_yticks([0, 1, 2, 3, 4, 5])
    ax.set_yticklabels(['N3 Sleep', 'N2 Sleep', 'N1 Sleep', 'REM Sleep', 'Wake', 'Abnormal'], )
    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_ylabel("Sleep Staging Result", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim([timestamp_num[0], timestamp_num[-1]])


if __name__ == '__main__':
    eeg_ = np.random.rand(30000)
    acc_ = np.random.rand(3, 30000)
    stage_res_ = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    plot_stage(eeg_, datetime.now(), 100, './eeg.eeg', acc_, stage_res_)
    plt.show()



"""
计算睡眠变量
"""
import numpy as np
import pandas as pd


def sl_waso_ar(hypno):
    sl = hypno.size
    gu_ = hypno.size
    hypno_mask = hypno != 0
    # 找连续10个epoch睡着的
    for i in range(hypno_mask.size):
        if np.sum(hypno_mask[i:i+10]) == 10:
            sl = i
            break
    # 从后找睡着的
    hypno_mask = hypno_mask[::-1]
    for j in range(hypno_mask.size):
        if np.sum(hypno_mask[j:j+1]) == 1:
            gu_ = hypno.size - j
            break
    # 睡与醒之间的清醒
    sleep_hypno= hypno[sl:gu_]
    sleep_hypno_mask = sleep_hypno == 0
    waso = np.sum(sleep_hypno_mask)
    # 清醒次数
    arr = sleep_hypno_mask
    merged_arr = [arr[0]]
    for i in range(1, len(arr)):
        if arr[i] != arr[i - 1]:
            merged_arr.append(arr[i])
    ar = np.sum(merged_arr)

    return sl, waso, ar


def sleep_metrics(eeg_path, hypno):
    # 总记录时间
    trt = hypno.shape[0]
    # 总睡眠时间
    tst = sum(1 for num in hypno if num in [1, 2, 3, 4]) if np.sum(hypno) > 0 else 0
    # 睡眠效率
    se = np.sum(hypno[hypno != 5] != 0) / hypno.size
    # 入睡后清醒时间， 入睡后清醒次数
    sol, waso, ar = sl_waso_ar(hypno) if tst > 10 else (0, 0, 0)

    n1 = np.sum(hypno == 1)
    n2 = np.sum(hypno == 2)
    n3 = np.sum(hypno == 3)
    rem = np.sum(hypno == 4)
    # save
    df = {
        "TRH(H)": [trt*30/3600],
        "TST(H)": [tst*30/3600],
        "SE(%)": [se*100],
        "SOL(H)": [sol*30/3600],
        "WASO(M)": [waso*30/3600],
        "AR": [ar],
        "Hypno": [hypno]
    }
    df = pd.DataFrame(df)
    save_path = eeg_path.replace('eeg.eeg', "analysis_results.xlsx")
    with pd.ExcelWriter(save_path) as writer:
        df.to_excel(writer, sheet_name='sleep variables')
    return trt, tst, se, sol, waso, ar, n1, n2, n3, rem


if __name__ == '__main__':
    hypno_ = np.array([0, 0, 0, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 1, 0, 1, 0, 1, 0])
    trt_, tst_, se_, sol_, waso_, ar_, n1_, n2_, n3_, rem_ = sleep_metrics('eeg.eeg', hypno_)
    print(trt_, tst_, se_, sol_, waso_, ar_, n1_, n2_, n3_, rem_)



import numpy as np
import scipy.signal as signal


def load_eeg_one_epoch(data_bytes, data_split):
    eeg, eeg_package_id, acc, acc_package_id = x8_epoch_data_parser(data_bytes, data_split[0], data_split[1])
    eeg = (eeg[0, :] - 32767) / 32767 * 2.5 * 1000 * 1000 / 192
    eeg = signal.resample(eeg, 100 * 30)
    return eeg, eeg_package_id


def int_from_bytes_8bit(byte_arr):
    buffer = np.asarray(
        [1, 256, np.power(np.int64(256), 2), np.power(np.int64(256), 3), np.power(np.int64(256), 4),
         np.power(np.int64(256), 5), np.power(np.int64(256), 6), np.power(np.int64(256), 7)])

    res = byte_arr[0] * buffer[0] + byte_arr[1] * buffer[1] + byte_arr[2] * buffer[2] + byte_arr[3] * buffer[
        3] + byte_arr[4] * buffer[4] + byte_arr[5] * buffer[5] + byte_arr[6] * buffer[6] + byte_arr[7] * buffer[7]
    return np.sum(res)


def int_from_bytes_4bit(byte_arr):
    buffer = np.asarray([1, 256, np.power(np.int64(256), 2), np.power(np.int64(256), 3)])

    res = byte_arr[0] * buffer[0] + byte_arr[1] * buffer[1] + byte_arr[2] * buffer[2] + byte_arr[3] * buffer[3]
    return np.sum(res)


def x8_epoch_data_parser(bytes_arr, eeg_length, acc_length):
    eeg_bytes = bytes_arr[0: + eeg_length]
    eeg, eeg_package_id = x8_epoch_eeg_or_acc_parser(eeg_bytes, 50, 2, 2)
    acc_bytes = bytes_arr[eeg_length:eeg_length + acc_length]
    acc, acc_package_id = x8_epoch_eeg_or_acc_parser(acc_bytes, 5, 2, 3)
    return eeg, eeg_package_id, acc, acc_package_id


def x8_epoch_eeg_or_acc_parser(eeg_or_acc_bytes, sample_count, point_bytes, channel_count):
    all_packages = np.array(eeg_or_acc_bytes)
    package_length = sample_count * point_bytes * channel_count + 18
    all_packages = all_packages.reshape(-1, package_length)

    all_package_id = all_packages[:, 10:14]
    all_package_data = all_packages[:, 18:package_length]

    all_package_id = all_package_id.astype(np.uint8)
    all_package_id = np.apply_along_axis(int_from_bytes_4bit, axis=1, arr=all_package_id)
    all_package_id = all_package_id.astype(np.int32)

    all_package_data = all_package_data.reshape(-1, point_bytes)
    all_package_data = all_package_data[:, 0] + all_package_data[:, 1] * 256

    data_total = all_package_data
    data_total = data_total[0:len(data_total) // channel_count * channel_count]
    data_total_T = data_total.reshape(-1, channel_count)
    data_total = []
    for i in range(channel_count):
        data_total.append(data_total_T[:, i])
    data_total = np.asarray(data_total)
    return data_total, all_package_id



"""
加载设备类型
"""


def load_device_type(eeg_path):
    eeg_data = open(eeg_path, 'rb')
    eeg_data.seek(12, 0)
    device_type = int.from_bytes(eeg_data.read(4), byteorder='little', signed=False)
    return device_type



import matplotlib.pyplot as plt
from lspopt import spectrogram_lspopt
import numpy as np
import matplotlib.dates as mdates
from datetime import timedelta, datetime
from matplotlib.colors import Normalize


def plot_spectrogram(ax, eeg, eeg_start_time, sf_eeg):
    sf = sf_eeg
    win_sample = int(30 * sf)
    assert eeg.size > 2 * win_sample, "`data` length must be at least 2 * `30`."
    f, t, Sxx = spectrogram_lspopt(eeg, sf, nperseg=win_sample, noverlap=0)
    t /= 3600
    timestamp = [eeg_start_time + timedelta(hours=hours) for hours in t]
    timestamp = mdates.date2num(timestamp)
    Sxx = 10 * np.log10(Sxx)  # Convert uV^2 / Hz --> dB / Hz

    v_min, v_max = np.percentile(Sxx, [0 + 5, 100 - 5])
    norm = Normalize(vmin=v_min, vmax=v_max)
    ax.pcolormesh(timestamp, f, Sxx, norm=norm, cmap='Spectral_r', antialiased=True, shading="auto")
    ax.xaxis_date()
    ax.set_ylim([0, 50])
    ax.set_yticks([5, 10, 15, 20, 25, 30, 35, 40, 45, 50])

    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_ylabel("Frequency [Hz]", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    return t


if __name__ == '__main__':
    eeg_ = np.random.rand(30000)
    fig, ax1 = plt.subplots(nrows=1, figsize=(12, 5))
    t_ = plot_spectrogram(ax1, eeg_, datetime.now(), 100)
    plt.show()



from qleegss.dataload.load_eeg_x8_one_night import load_eeg_x8_one_night
from qleegss.dataload.load_acc import load_acc
from qleegss.dataload.load_sti import load_sti
from qleegss.model.predict import predict_one_trail
from qleegss.plot.plot_sw import plot_sw
from qleegss.plot.plot_stage import plot_stage
from qleegss.other.metric import sleep_metrics
from qleegss.pdf.sleep_report import generate_sleep_report
import logging
from qleegss.dataload.load_device_type import load_device_type
from qleegss.dataload.load_eeg_mr4_one_night import load_eeg_mr4_one_night
from qleegss.plot.plot_data_preview import plot_preview_mr4
from qleegss.pdf.data_preview import generate_data_preview
from qleegss.dataload.load_eeg_mr2_one_night import load_eeg_mr2_one_night
from qleegss.plot.plot_data_preview import plot_preview_mr2


class DataHandler:
    def __init__(self, eeg_path_p=None, acc_path_p=None, sti_path_p=None):
        self.eeg = None
        self.start_time = None
        self.eeg_path = eeg_path_p
        self.sf_eeg = 100
        self.acc = None
        self.acc_path = acc_path_p
        self.sti = None
        self.sti_path = sti_path_p
        self.stage_result = None
        self.phone = None
        self.name = None
        self.end_time = None
        self.disconnect_rate = None
        self.package_loss_rate = None
        self.device = None
        self.ecg = None
        self.emg = None
        self.eeg0 = None
        self.eeg1 = None
        self.header_mac = None
        self.box_mac = None

    def load_eeg_x8_one_night(self):
        self.eeg, self.start_time, self.end_time,  self.disconnect_rate, self.package_loss_rate = load_eeg_x8_one_night(self.eeg_path)

    def load_eeg_mr4_one_night(self):
        self.ecg, self.emg, self.eeg0, self.eeg1, self.start_time, self.end_time, self.disconnect_rate, self.package_loss_rate, \
            self.header_mac, self.box_mac = load_eeg_mr4_one_night(self.eeg_path)

    def load_eeg_mr2_one_night(self):
        self.eeg0, self.eeg1, self.start_time, self.end_time, self.disconnect_rate, self.package_loss_rate, \
            self.header_mac, self.box_mac = load_eeg_mr2_one_night(self.eeg_path)

    def load_acc(self):
        self.acc = load_acc(self.acc_path)

    def load_sti(self):
        self.sti = load_sti(self.sti_path)

    def load_device_type(self):
        self.device = load_device_type(self.eeg_path)

    def sleep_stage(self):
        self.stage_result = predict_one_trail(self.eeg)

    def clear_eeg(self):
        self.eeg = None

    def clear_acc(self):
        self.acc = None

    def clear_sti(self):
        self.sti = None

    def clear_mr_data(self):
        self.ecg = None
        self.emg = None
        self.eeg0 = None
        self.eeg1 = None


if __name__ == '__main__':
    logging.basicConfig(filename='example.log', level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    eeg_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\eeg.eeg'
    acc_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\acc.acc'
    sti_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\sti.log'
    try:
        data = DataHandler(eeg_path, acc_path, sti_path)
        data.load_device_type()

        if data.device == 101:
            logging.info('--------- X8 device ---------')
            data.load_eeg_x8_one_night() if data.eeg_path is not None else None
            data.load_acc() if data.acc_path is not None else None
            data.load_sti() if data.sti_path is not None else None
            logging.info('load data success.')
            # 慢波图绘制
            plot_sw(data.eeg, data.start_time, data.sf_eeg, data.eeg_path, data.sti) if data.sti is not None else None
            logging.info('slow wave success.')
            # 清除sti
            data.clear_sti()
            # 生成分期
            data.sleep_stage()
            logging.info('sleep stage success.')
            # 绘制分期
            plot_stage(data.eeg, data.start_time, data.sf_eeg, data.eeg_path, data.acc, data.stage_result)
            logging.info('plot stage success.')
            # 清数据
            data.clear_eeg()
            data.clear_acc()
            # 计算睡眠指标&保存成excel
            sleep_metrics(data.eeg_path, data.stage_result)
            logging.info('save xlsx success.')
            # pdf
            generate_sleep_report(data)
            logging.info('save report success.')
        # 4通道小鼠
        elif data.device == 102:
            logging.info('--------- MR4 device ---------')
            # 加载数据
            data.load_eeg_mr4_one_night() if data.eeg_path is not None else None
            data.load_acc() if data.acc_path is not None else None
            logging.info('load data success.')
            # data preview
            plot_preview_mr4(data.eeg0, data.eeg1, data.ecg, data.emg, data.acc, data.start_time, data.sf_eeg, data.eeg_path)
            logging.info('plot data success.')
            data.clear_mr_data()
            # save pdf
            generate_data_preview(data)
            logging.info('save data preview success.')

        # 2通道小鼠
        elif data.device == 44:
            logging.info('--------- MR2 device ---------')
            # 加载数据
            data.load_eeg_mr2_one_night() if data.eeg_path is not None else None
            data.load_acc() if data.acc_path is not None else None
            logging.info('load data success.')
            # data preview
            plot_preview_mr2(data.eeg0, data.eeg1, data.acc, data.start_time, data.sf_eeg, data.eeg_path)
            logging.info('plot data success.')
            data.clear_mr_data()
            # save pdf
            generate_data_preview(data)
            logging.info('save data preview success.')

        else:
            logging.info('Unknown Device Type')

    except Exception as e:
        logging.error(f"{e}")



import os
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus import Table, TableStyle, Image


def generate_data_preview(data):
    eeg_path = data.eeg_path
    fig_path = eeg_path.replace('eeg.eeg', 'data_preview.png')
    pdf_path = eeg_path.replace('eeg.eeg', 'data_preview.pdf')
    record_date = data.start_time.strftime('%Y-%m-%d')

    content = {
        "sleep_plot": fig_path,
        "name": data.name,
        "phone_number": data.phone,
        "record_date": record_date,
        "record_start_time": data.start_time,
        "record_end_time": data.end_time,
        "disconnection_rate": "{:.2f}%".format(data.disconnect_rate),
        "package_loss_rate": "{:.2f}%".format(data.package_loss_rate),
        "header_mac": data.header_mac,
        "box_mac": data.box_mac,
    }

    pdf_plot(pdf_path, content)


def pdf_plot(pdf_save_path, Results):
    """
    生成分析报告
    :param pdf_save_path: PDF文件保存路径
    :param Results: 结果集
    :return:
    """
    base_path = os.path.abspath(__file__)
    dir_path = os.path.dirname(base_path)
    font_path = os.path.join(dir_path, "msyhbd.ttc")
    pdfmetrics.registerFont(TTFont("Yahei", font_path))

    pdf_file = Canvas(pdf_save_path, pagesize=A4)

    TFimg = Image(Results["sleep_plot"], width=495, height=420) if Results["sleep_plot"] is not None else Image(width=495, height=420)

    Description = "备注：\n" \
                  "断连率：因记录子与接收器距离太远导致数据无法记录的时间占总记录时间的比例，断连率大于5%可能对分析结果造成明显影响。\n" \
                  "丢包率：因障碍物阻挡导致通信质量较差时导致的丢包。正常情况下，丢包率小于0.5%，丢包率大于5%可能对分析结果造成明显影响。"

    name = Results.get("name") if Results.get("name") is not None else "--"
    phone_number = Results.get("phone_number") if Results.get("phone_number") is not None else "--"
    record_date = Results.get("record_date") if Results.get("record_date") is not None else "--"
    record_start_time = Results.get("record_start_time") if Results.get("record_start_time") is not None else "--"
    record_end_time = Results.get("record_end_time") if Results.get("record_end_time") is not None else "--"
    package_loss_rate = Results.get("package_loss_rate") if Results.get("package_loss_rate") is not None else "--"
    disconnection_rate = Results.get("disconnection_rate") if Results.get("disconnection_rate") is not None else "--"
    sl = Results.get("sl") if Results.get("sl") is not None else None
    if sl is not None:
        sl = "{:.2f}".format(sl) if sl > 0 else "--"

    header_mac = Results.get("header_mac") if Results.get("header_mac") is not None else "--"
    box_mac = Results.get("box_mac") if Results.get("box_mac") is not None else "--"

    data = [
        ["数据预览"],
        ["姓名", name, "手机号", phone_number, "监测日期", record_date],

        ["记录子MAC", header_mac, '', "接收器MAC", box_mac],

        ["记录时间"],
        ["记录开始时间", record_start_time, "", "记录结束时间", record_end_time],
        ["信号质量"],
        ["断连率(%)", disconnection_rate, "", "丢包率(%)", package_loss_rate],
        ["数据预览图"],
        [TFimg],
        [Description],
        [""]
    ]
    SPAN_SPAN_SPAN_SPAN_ = [("SPAN", (0, 0), (-1, 0)),
                            ("SPAN", (0, 3), (-1, 3)),
                            ("SPAN", (0, 5), (-1, 5)),

                            ("SPAN", (0, 7), (-1, 7)),
                            ("SPAN", (0, 8), (-1, 8)),
                            ("SPAN", (0, 9), (-1, 9)),

                            ("SPAN", (1, 2), (2, 2)),
                            ("SPAN", (4, 2), (5, 2)),

                            ("SPAN", (1, 4), (2, 4)),
                            ("SPAN", (-2, 4), (-1, 4)),

                            ("SPAN", (1, 6), (2, 6)),
                            ("SPAN", (4, 6), (5, 6)),

                            ("SPAN", (0, -1), (-1, -1)),

                            ]
    merged_cells = SPAN_SPAN_SPAN_SPAN_

    table = Table(data, colWidths=[80, 100, 80, 100, 80, 100])
    table.setStyle(TableStyle([
        *merged_cells,
        ("BACKGROUND", (0, 0), (-1, 0), colors.grey),
        ("TEXTCOLOR", (0, 0), (-1, 0), colors.whitesmoke),
        ("ALIGN", (0, 0), (-2, -3), "CENTER"),
        ("ALIGN", (-1, -2), (-1, -2), "TA_LEFT"),
        ("FONTNAME", (0, 0), (-1, -1), "Yahei"),
        ("FONTSIZE", (0, 0), (-1, -1), 8),
        ("BOTTOMPADDING", (0, 0), (-1, 0), 5),
        ("TOPPADDING", (0, 0), (-1, 0), 5),
        ("GRID", (0, 0), (-1, -2), 1, colors.black),
        ("BOTTOMPADDING", (0, -1), (-1, -1), 100),
    ]))

    table.wrapOn(pdf_file, 0, 0)
    table.drawOn(pdf_file, 28, 10)

    pdf_file.save()



import pandas as pd
import numpy as np
import os
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus import Table, TableStyle, Image


def generate_sleep_report(data):
    eeg_path = data.eeg_path
    fig_path = eeg_path.replace('eeg.eeg', 'sleep_fig.png')
    xlsx_path = eeg_path.replace('eeg.eeg', 'analysis_results.xlsx')
    pdf_path = eeg_path.replace('eeg.eeg', 'sleep_report.pdf')
    record_date = data.start_time.strftime('%Y-%m-%d')
    df = pd.read_excel(xlsx_path, engine='openpyxl')

    Hypno = data.stage_result

    n1_n2 = 30*(np.sum(Hypno == 1) + np.sum(Hypno == 2)) / 60
    n3 = 30*np.sum(Hypno == 3) / 60
    rem = 30*np.sum(Hypno == 4) / 60

    content = {
        "sleep_plot": fig_path,
        "name": data.name,
        "phone_number": data.phone,
        "record_date": record_date,
        "record_start_time": data.start_time,
        "record_end_time": data.end_time,
        "disconnection_rate": "{:.2f}%".format(data.disconnect_rate),
        "package_loss_rate": "{:.2f}%".format(data.package_loss_rate),
        "trt": df['TRH(H)'].values[0],
        "tst": df['TST(H)'].values[0],
        "se": df['SE(%)'].values[0],
        "sl": df['SOL(H)'].values[0],
        "waso": df['WASO(M)'].values[0],
        "ar": df["AR"].values[0],
        "N1/N2": n1_n2,
        "N3": n3,
        "REM": rem
    }

    pdf_plot(pdf_path, content)


def pdf_plot(pdf_save_path, Results):
    """
    生成睡眠分析报告
    :param pdf_save_path: PDF文件保存路径
    :param Results: 结果集
    :return:
    """
    base_path = os.path.abspath(__file__)
    dir_path = os.path.dirname(base_path)
    font_path = os.path.join(dir_path, "msyhbd.ttc")
    pdfmetrics.registerFont(TTFont("Yahei", font_path))

    pdf_file = Canvas(pdf_save_path, pagesize=A4)

    TFimg = Image(Results["sleep_plot"], width=495, height=420) if Results["sleep_plot"] is not None else Image(
        width=495, height=420)
    Description = "备注：\n" \
                  "断连率：因额贴与设备基座距离太远导致数据无法记录的时间占总记录时间的比例，大于5%可能对睡眠分析结果造成明显影响。\n" \
                  "丢包率：因夜间用户移动超出额贴与通信距离导致的蓝牙信号传输中断所带来的丢包。丢包率通常小于0.5%，大于5%可对睡眠指标造成明显影响。\n" \
                  "总记录时间（TRT）：从关灯到开灯的时间(先用开始记录/停止时间替代)，是睡眠记录的全部时长。\n" \
                  "总睡眠时间（TST）：关灯至开灯时间内实际睡眠时间总和，即各睡眠期（N1期，N2期，N3期，R期）时间的总和。\n" \
                  "入睡延迟（SOL）：从关灯（开始记录）到出现第一帧睡眠期的时间。\n" \
                  "入睡后清醒时间（WASO）/次数（AR）：从稳定的睡眠期到睡眠结束（最后一个睡眠周期）之间，所有的清醒时间的总和/清醒次数。\n" \
                  "睡眠效率（SE）：总睡眠时间/总记录时间×100％。\n" \
                  "各睡眠期比例：各睡眠期（N1/N2期，N3期，R期）分别累计的时间占总睡眠时间（不包括清醒时间）的百分比。\n" \
                  "睡眠分期图中，Abnormal一般表示未佩戴状态（接触不良或者脱落）或者存在较大干扰导致信号异常。"
    n3_time = Results.get("N3") if Results.get("N3") is not None else None
    n1n2_time = Results.get("N1/N2") if Results.get("N3") is not None else None
    rem_time = Results.get("REM") if Results.get("N3") is not None else None
    n3_rate = None
    n1n2_rate = None
    rem_rate = None
    if n3_time is not None and n1n2_time is not None and rem_time is not None:
        n3_time = n3_time if n3_time > 0 else 0
        n1n2_time = n1n2_time if n1n2_time > 0 else 0
        rem_time = rem_time if rem_time > 0 else 0
        total_time = n3_time + n1n2_time + rem_time
        if total_time == 0:
            n3_rate = "--"
            n1n2_rate = "--"
            rem_rate = "--"
        else:
            n3_rate = "{:.2f}%".format(n3_time / total_time * 100)
            n1n2_rate = "{:.2f}%".format(n1n2_time / total_time * 100)
            rem_rate = "{:.2f}%".format(rem_time / total_time * 100)
    name = Results.get("name") if Results.get("name") is not None else "--"
    phone_number = Results.get("phone_number") if Results.get("phone_number") is not None else "--"
    record_date = Results.get("record_date") if Results.get("record_date") is not None else "--"
    record_start_time = Results.get("record_start_time") if Results.get("record_start_time") is not None else "--"
    record_end_time = Results.get("record_end_time") if Results.get("record_end_time") is not None else "--"
    package_loss_rate = Results.get("package_loss_rate") if Results.get("package_loss_rate") is not None else "--"
    disconnection_rate = Results.get("disconnection_rate") if Results.get("disconnection_rate") is not None else "--"
    trt = "{:.2f}".format(Results.get("trt")) if Results.get("trt") is not None else "--"
    tst = "{:.2f}".format(Results.get("tst")) if Results.get("tst") is not None else "--"
    se = Results.get("se") if Results.get("se") is not None else "--"
    sl = Results.get("sl") if Results.get("sl") is not None else "--"
    if sl is not None:
        sl = "{:.2f}".format(sl) if sl > 0 else "--"

    data = [
        ["居家睡眠监测报告"],
        ["姓名", name, "手机号", phone_number, "监测日期", record_date],

        ["记录时间"],
        ["记录开始时间", record_start_time, "", "记录结束时间", record_end_time],
        ["信号质量"],
        ["断连率(%)", disconnection_rate, "", "丢包率(%)", package_loss_rate, ""],
        ["睡眠综合分析图"],
        [TFimg],
        ["睡眠参数"],
        ["总记录时间(TRT,min)", trt, "", "总睡眠时间(TST,min)", tst],
        ["睡眠效率(SE,%)", se, "", "入睡延迟(SOL,min)", sl],
        ["入睡后清醒时间(WASO,min)", "{:.2f}".format(Results.get("waso")) if Results.get("waso") is not None else "--",
         "",
         "入睡后清醒次数(AR)", Results.get("ar") if Results.get("ar") is not None else "--"],

        ["各期睡眠时间(min)"],
        ["N3", n3_time, "N1/N2", n1n2_time, "REM", rem_time],
        ["各期睡眠比例(%)"],
        ["N3", n3_rate if n3_rate is not None else "--", "N1/N2", n1n2_rate if n1n2_rate is not None else "--", "REM",
         rem_rate if rem_rate is not None else "--"],
        [Description]
    ]
    SPAN_SPAN_SPAN_SPAN_ = [("SPAN", (0, 0), (-1, 0)),
                            ("SPAN", (0, 2), (-1, 2)),
                            ("SPAN", (0, 4), (-1, 4)),

                            ("SPAN", (0, 6), (-1, 6)),
                            ("SPAN", (0, 7), (-1, 7)),
                            ("SPAN", (0, 8), (-1, 8)),

                            # ("SPAN", (0, 10), (-1, 10)),
                            ("SPAN", (0, 12), (-1, 12)),
                            ("SPAN", (0, 14), (-1, 14)),
                            ("SPAN", (0, 16), (-1, 16)),

                            ("SPAN", (1, 3), (2, 3)),
                            ("SPAN", (-2, 3), (-1, 3)),

                            ("SPAN", (1, 5), (2, 5)),
                            ("SPAN", (-2, 5), (-1, 5)),

                            ("SPAN", (1, 9), (2, 9)),
                            ("SPAN", (-2, 9), (-1, 9)),

                            ("SPAN", (1, 10), (2, 10)),
                            ("SPAN", (-2, 10), (-1, 10)),

                            ("SPAN", (1, 11), (2, 11)),
                            ("SPAN", (-2, 11), (-1, 11)),

                            ]
    merged_cells = SPAN_SPAN_SPAN_SPAN_
    # print("------------Creating Span---------------------")
    table = Table(data)
    table.setStyle(TableStyle([
        *merged_cells,
        ("BACKGROUND", (0, 0), (-1, 0), colors.grey),
        ("TEXTCOLOR", (0, 0), (-1, 0), colors.whitesmoke),
        ("ALIGN", (0, 0), (-2, -2), "CENTER"),
        ("ALIGN", (-1, -1), (-1, -1), "TA_LEFT"),
        ("FONTNAME", (0, 0), (-1, -1), "Yahei"),
        ("FONTSIZE", (0, 0), (-1, -1), 8),
        ("BOTTOMPADDING", (0, 0), (-1, 0), 5),
        ("TOPPADDING", (0, 0), (-1, 0), 5),
        ("GRID", (0, 0), (-1, -1), 1, colors.black),
    ]))
    # print("------------Creating Table---------------------")
    # 将表格绘制在指定位置
    table.wrapOn(pdf_file, 0, 0)
    table.drawOn(pdf_file, 28, 10)
    # 保存并关闭 PDF 文件
    pdf_file.save()



import numpy as np
import torch
import os
from qleegss.model.sleepyco_net import MainModel
from tqdm import tqdm


def predict_one_trail(eeg):
    net = create_model()
    res = []
    net.eval()
    eeg = torch.from_numpy(eeg.reshape(-1, 100*30)).float()
    with torch.no_grad():
        for inputs in tqdm(eeg):
            inputs = inputs.reshape(1, 1, -1)
            outputs = net(inputs)
            outputs_sum = torch.zeros_like(outputs[0])

            for j in range(len(outputs)):
                outputs_sum += outputs[j]
            predicted = torch.argmax(outputs_sum)
            res.append(predicted.item())
    return np.asarray(res)


def predict_one_epoch(eeg):
    net = create_model()
    net.eval()
    eeg = torch.from_numpy(eeg).reshape(1, 1, -1).float()
    with torch.no_grad():
        outputs = net(eeg)
        outputs_sum = torch.zeros_like(outputs[0])
        for j in range(len(outputs)):
            outputs_sum += outputs[j]
        predicted = torch.argmax(outputs_sum)
        map_dict = {
            3: 0,
            2: 1,
            1: 2,
            4: 3,
            0: 4,
            5: 5
        }
    return map_dict.get(int(predicted.item()))


def create_model():
    script_path = os.path.abspath(__file__)
    script_dir = os.path.dirname(script_path)
    net = MainModel()
    net.eval()
    state_dict = torch.load(script_dir + '/ckpt_fold-01.pth', map_location=torch.device('cpu'))
    new_state_dict = {}
    for k, v in state_dict.items():
        name = k.replace('module.', '')
        new_state_dict[name] = v
    net.load_state_dict(new_state_dict)
    return net


if __name__ == '__main__':
    net_ = create_model()
    print(net_(torch.rand((1, 1, 30*250))))



import torch.nn as nn
import torch.nn.functional as F
import torch
import math


class SleePyCoBackbone(nn.Module):

    def __init__(self):
        super(SleePyCoBackbone, self).__init__()

        # architecture
        self.init_layer = self.make_layers(in_channels=1, out_channels=64, n_layers=2, maxpool_size=None, first=True)
        self.layer1 = self.make_layers(in_channels=64, out_channels=128, n_layers=2, maxpool_size=5)
        self.layer2 = self.make_layers(in_channels=128, out_channels=192, n_layers=3, maxpool_size=5)
        self.layer3 = self.make_layers(in_channels=192, out_channels=256, n_layers=3, maxpool_size=5)
        self.layer4 = self.make_layers(in_channels=256, out_channels=256, n_layers=3, maxpool_size=5)

        self.fp_dim = 128
        self.conv_c5 = nn.Conv1d(256, self.fp_dim, 1, 1, 0)

        self.conv_c4 = nn.Conv1d(256, self.fp_dim, 1, 1, 0)

        self.conv_c3 = nn.Conv1d(192, self.fp_dim, 1, 1, 0)

    def make_layers(self, in_channels, out_channels, n_layers, maxpool_size, first=False):
        layers = []
        layers = layers + [MaxPool1d(maxpool_size)] if not first else layers

        for i in range(n_layers):
            conv1d = nn.Conv1d(in_channels, out_channels, kernel_size=3, padding=1)
            layers += [conv1d, nn.BatchNorm1d(out_channels)]
            if i == n_layers - 1:
                layers += [ChannelGate(in_channels)]
            layers += [nn.PReLU()]
            in_channels = out_channels

        return nn.Sequential(*layers)

    def forward(self, x):
        out = []

        c1 = self.init_layer(x)
        c2 = self.layer1(c1)
        c3 = self.layer2(c2)
        c4 = self.layer3(c3)
        c5 = self.layer4(c4)

        p5 = self.conv_c5(c5)
        out.append(p5)
        p4 = self.conv_c4(c4)
        out.append(p4)
        p3 = self.conv_c3(c3)
        out.append(p3)

        return out


class MaxPool1d(nn.Module):
    def __init__(self, maxpool_size):
        super(MaxPool1d, self).__init__()
        self.maxpool_size = maxpool_size
        self.maxpool = nn.MaxPool1d(kernel_size=maxpool_size, stride=maxpool_size)

    def forward(self, x):
        _, _, n_samples = x.size()
        if n_samples % self.maxpool_size != 0:
            pad_size = self.maxpool_size - (n_samples % self.maxpool_size)
            if pad_size % 2 != 0:
                left_pad = pad_size // 2
                right_pad = pad_size // 2 + 1
            else:
                left_pad = pad_size // 2
                right_pad = pad_size // 2
            x = F.pad(x, (left_pad, right_pad), mode='constant')

        x = self.maxpool(x)

        return x


class BasicConv(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride=1, padding=0, dilation=1, groups=1, relu=True,
                 bn=True, bias=False):
        super(BasicConv, self).__init__()
        self.out_channels = out_planes
        self.conv = nn.Conv1d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding,
                              dilation=dilation, groups=groups, bias=bias)
        self.bn = nn.BatchNorm1d(out_planes, eps=1e-5, momentum=0.01, affine=True) if bn else None
        self.relu = nn.ReLU() if relu else None

    def forward(self, x):
        x = self.conv(x)
        if self.bn is not None:
            x = self.bn(x)
        if self.relu is not None:
            x = self.relu(x)
        return x


class ChannelGate(nn.Module):
    def __init__(self, gate_channels, reduction_ratio=16, pool_types=['avg']):
        super(ChannelGate, self).__init__()
        self.gate_channels = gate_channels
        self.mlp = nn.Sequential(
            nn.Flatten(),
            nn.Linear(gate_channels, gate_channels // reduction_ratio),
            nn.ReLU(),
            nn.Linear(gate_channels // reduction_ratio, gate_channels)
        )
        self.pool_types = pool_types

    def forward(self, x):
        channel_att_sum = None
        for pool_type in self.pool_types:
            if pool_type == 'avg':
                avg_pool = F.avg_pool1d(x, x.size(2), stride=x.size(2))
                channel_att_raw = self.mlp(avg_pool)
            elif pool_type == 'max':
                max_pool = F.max_pool1d(x, x.size(2), stride=x.size(2))
                channel_att_raw = self.mlp(max_pool)
            elif pool_type == 'lp':
                lp_pool = F.lp_pool2d(x, 2, (x.size(2), x.size(3)), stride=(x.size(2), x.size(3)))
                channel_att_raw = self.mlp(lp_pool)
            elif pool_type == 'lse':
                # LSE pool only
                lse_pool = logsumexp_2d(x)
                channel_att_raw = self.mlp(lse_pool)

            if channel_att_sum is None:
                channel_att_sum = channel_att_raw
            else:
                channel_att_sum = channel_att_sum + channel_att_raw

        scale = F.sigmoid(channel_att_sum).unsqueeze(2).expand_as(x)
        return x * scale


def logsumexp_2d(tensor):
    tensor_flatten = tensor.view(tensor.size(0), tensor.size(1), -1)
    s, _ = torch.max(tensor_flatten, dim=2, keepdim=True)
    outputs = s + (tensor_flatten - s).exp().sum(dim=2, keepdim=True).log()
    return outputs


feature_len_dict = [[5, 24, 120], [10, 48, 240], [15, 72, 360], [20, 96, 480], [24, 120, 600], [29, 144, 720],
                    [34, 168, 840], [39, 192, 960], [44, 216, 1080], [48, 240, 1200]]


class PositionalEncoding(nn.Module):

    def __init__(self, in_features, out_features):
        super(PositionalEncoding, self).__init__()
        self.num_scales = 128

        self.fc = nn.Linear(in_features=in_features, out_features=out_features)
        self.act_fn = nn.PReLU()

        if self.num_scales > 1:
            self.max_len = feature_len_dict[10 - 1][3 - 1]
        else:
            self.max_len = 5000

        pe = torch.zeros(self.max_len, out_features)
        position = torch.arange(0, self.max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, out_features, 2).float() * (-math.log(10000.0) / out_features))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = self.act_fn(self.fc(x))

        if self.num_scales > 1:
            hop = self.max_len // x.size(0)
            pe = self.pe[hop // 2::hop, :]
        else:
            pe = self.pe

        if pe.shape[0] != x.size(0):  # 保证pe大小和x的大小相同
            pe = pe[:x.size(0), :]

        x = x + pe

        return x


class Transformer(nn.Module):

    def __init__(self, nheads, num_encoder_layers, pool='attn'):

        super(Transformer, self).__init__()

        self.model_dim = 128
        self.feedforward_dim = 128

        self.in_features = 128
        self.out_features = 128

        self.pos_encoding = PositionalEncoding(self.in_features, self.out_features)

        self.transformer_layer = nn.TransformerEncoderLayer(
            d_model=self.model_dim,
            nhead=nheads,
            dim_feedforward=self.feedforward_dim,
            dropout=0
        )
        self.transformer = nn.TransformerEncoder(self.transformer_layer, num_layers=num_encoder_layers)

        self.pool = pool

        if pool == 'attn':
            self.w_ha = nn.Linear(self.model_dim, self.model_dim, bias=True)
            self.w_at = nn.Linear(self.model_dim, 1, bias=False)

        self.fc = nn.Linear(self.model_dim, 6)

    def forward(self, x):
        x = x.transpose(0, 1)
        x = self.pos_encoding(x)
        x = self.transformer(x)
        x = x.transpose(0, 1)

        if self.pool == 'mean':
            x = x.mean(dim=1)
        elif self.pool == 'last':
            x = x[:, -1]
        elif self.pool == 'attn':
            a_states = torch.tanh(self.w_ha(x))
            alpha = torch.softmax(self.w_at(a_states), dim=1).view(x.size(0), 1, x.size(1))
            x = torch.bmm(alpha, a_states).view(x.size(0), -1)
        elif self.pool == None:
            x = x
        else:
            raise NotImplementedError

        out = self.fc(x)

        return out


class MainModel(nn.Module):

    def __init__(self):

        super(MainModel, self).__init__()

        self.feature = SleePyCoBackbone()

        self.classifier = Transformer(nheads=8, num_encoder_layers=6, pool='attn')

    def get_max_len(self, features):
        len_list = []
        for feature in features:
            len_list.append(feature.shape[1])

        return max(len_list)

    def forward(self, x):
        outputs = []
        features = self.feature(x)

        for feature in features:
            feature = feature.transpose(1, 2)
            output = self.classifier(feature)
            outputs.append(output)

        return outputs



"""
load eeg x8 data
"""
import numpy as np
from datetime import datetime
import scipy


def load_eeg_x8_one_night(eeg_path):
    # eeg data
    eeg_data = open(eeg_path, 'rb')
    eeg_len = len(eeg_data.read())
    eeg_data.seek(8, 0)
    length = int.from_bytes(eeg_data.read(4), byteorder='little', signed=False)
    eeg_data.seek(length, 0)
    all_package = np.array(list(eeg_data.read(eeg_len - length)))
    eeg_data.seek(21, 0)
    sample_count = int.from_bytes(eeg_data.read(4), byteorder='little', signed=False)
    eeg_data.seek(16, 0)
    point_bytes = int.from_bytes(eeg_data.read(1), byteorder='little', signed=False)
    channel_count = 2  # Note: 2
    package_length = sample_count * point_bytes * channel_count + 18
    all_package = all_package.reshape(-1, package_length)
    all_package_data = all_package[:, 18:package_length]
    all_package_data = all_package_data.reshape(-1, point_bytes)
    all_package_data = all_package_data[:, 0] + all_package_data[:, 1] * 256
    raw_data = np.squeeze(all_package_data)
    data = np.transpose(raw_data.reshape(-1, channel_count))
    data = data[0, :]
    eeg = (data - 32767) / 32767 * 2.5 * 1000 * 1000 / 192
    # start time
    eeg_data.seek(33, 0)
    start_time = int.from_bytes(eeg_data.read(8), byteorder='little', signed=False) / 1000
    eeg_start_time = datetime.fromtimestamp(start_time)
    # end_time
    eeg_data.seek(41, 0)
    end_time = int.from_bytes(eeg_data.read(8), byteorder='little', signed=False) / 1000
    eeg_end_time = datetime.fromtimestamp(end_time)

    eeg = eeg[:int(eeg.shape[0] / (30 * 500)) * (30 * 500)].reshape(-1, 30 * 500)
    eeg = scipy.signal.resample(eeg, 100 * 30, axis=1).ravel()

    # 断连和丢包率
    all_package_time = all_package[:, 2:10]
    all_package_id = all_package[:, 10:14]

    # computing package loss rate/disconnection rate
    all_package_id = all_package_id.astype(np.uint8)
    all_package_id = np.apply_along_axis(int_from_bytes_4bit, axis=1, arr=all_package_id)
    all_package_id = all_package_id.astype(np.int32)

    all_package_time = all_package_time.astype(np.uint8)
    all_package_time = np.apply_along_axis(int_from_bytes_8bit, axis=1, arr=all_package_time)
    all_package_time = all_package_time.astype(np.int32)

    package_time_interval = all_package_time[1:] - all_package_time[:-1]
    disconnect_point = np.where(package_time_interval > 3000)[0]

    disconnection_sum = 0

    package_segment = [0]
    if disconnect_point is not None and len(disconnect_point) > 0:
        for i in range(len(disconnect_point)):
            package_segment.append(disconnect_point[i])
            package_segment.append(disconnect_point[i] + 1)
            disconnection_sum += all_package_time[disconnect_point[i] + 1] - all_package_time[disconnect_point[i]]

    disconnect_rate = disconnection_sum / all_package_time[-1]

    package_segment.append(all_package_id.shape[0] - 1)
    package_segment = np.array(package_segment).reshape([-1, 2])

    package_sum = 0
    loss_package_sum = 0
    for i in range(package_segment.shape[0]):
        left = package_segment[i][0]
        right = package_segment[i][1]
        package_sum += all_package_id[right] - 0 + 1
        loss_package_sum += all_package_id[right] - 0 + 1 - (right - left + 1)
    package_loss_rate = loss_package_sum / package_sum

    return eeg, eeg_start_time, eeg_end_time, disconnect_rate, package_loss_rate


def int_from_bytes_8bit(byte_arr):
    buffer = np.asarray(
        [1, 256, np.power(np.int64(256), 2), np.power(np.int64(256), 3), np.power(np.int64(256), 4),
         np.power(np.int64(256), 5), np.power(np.int64(256), 6), np.power(np.int64(256), 7)])

    res = byte_arr[0] * buffer[0] + byte_arr[1] * buffer[1] + byte_arr[2] * buffer[2] + byte_arr[3] * buffer[
        3] + byte_arr[4] * buffer[4] + byte_arr[5] * buffer[5] + byte_arr[6] * buffer[6] + byte_arr[7] * buffer[7]
    return np.sum(res)


def int_from_bytes_4bit(byte_arr):
    buffer = np.asarray([1, 256, np.power(np.int64(256), 2), np.power(np.int64(256), 3)])
    res = byte_arr[0] * buffer[0] + byte_arr[1] * buffer[1] + byte_arr[2] * buffer[2] + byte_arr[3] * buffer[3]
    return np.sum(res)



"""
加载acc数据
"""
import numpy as np


def load_acc(acc_path):
    acc_data = open(acc_path, 'rb')
    acc_len = len(acc_data.read())
    acc_data.seek(8, 0)
    length = int.from_bytes(acc_data.read(4), byteorder='little', signed=False)
    acc_data.seek(length, 0)
    all_package = np.array(list(acc_data.read(acc_len - length)))
    acc_data.seek(21, 0)
    sample_count = int.from_bytes(acc_data.read(4), byteorder='little', signed=False)
    acc_data.seek(16, 0)
    point_bytes = int.from_bytes(acc_data.read(1), byteorder='little', signed=False)
    channel_count = 3  # NOTE: 3通道
    package_length = sample_count * point_bytes * channel_count + 18
    all_package = all_package.reshape(-1, package_length)
    all_package_data = all_package[:, 18:package_length]
    all_package_data = all_package_data.reshape(-1, point_bytes)
    all_package_data = all_package_data[:, 0] + all_package_data[:, 1] * 256
    raw_data = np.squeeze(all_package_data)
    data = np.transpose(raw_data.reshape(-1, channel_count))
    acc = data - 32767
    acc = acc[:, : (acc.shape[1] // (15 * 50)) * (15 * 50)]
    return acc



from qleegss.handler import DataHandler


if __name__ == '__main__':
    eeg_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\eeg.eeg'
    acc_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\acc.acc'
    sti_path = r'D:\pythonProject\sleep_stage\dev_test_data\13592029172_0321-22_16_05_0322-08_04_45_2\sti.log'
    data = DataHandler(eeg_path, acc_path, sti_path)
    data.load_sti()



from setuptools import setup, find_packages

setup(
    name='qleegss',
    version='0.0.6',
    keywords='eeg sleep staging analysis',
    description='a python analyse tool for LM Data Recorder data',
    license='MIT License',
    url='https://github.com/eegion/qleegss',
    author='zhd',
    author_email='shrugor@163.com',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'my_package': ['ckpt_fold-01.pth', 'msyhbd.ttc']
    },
    platforms='any',
    install_requires=['scipy', 'numpy', 'torch', 'tqdm', 'lspopt', 'pandas', 'reportlab', 'openpyxl', 'matplotlib'],
)



"""
绘制数据预览
"""
from qleegss.plot.plot_acc import plot_acc
import matplotlib.pyplot as plt
from qleegss.plot.plot_eeg_spectrogram import plot_spectrogram
import numpy as np
import matplotlib.dates as mdates
from datetime import timedelta


def plot_preview_mr4(eeg0, eeg1, ecg, emg, acc, eeg_start_time, sf_eeg, eeg_path):
    fig, (ax1, ax2, ax3, ax4, ax5) = plt.subplots(5, 1, figsize=(20, 5 * 4))
    # spectrogram
    t = plot_spectrogram(ax1, eeg0, eeg_start_time, sf_eeg)
    t = plot_spectrogram(ax2, eeg1, eeg_start_time, sf_eeg)
    # ecg
    plot_ecg(ax3, ecg, eeg_start_time)
    # emg
    plot_emg(ax4, emg, eeg_start_time)
    # acc
    plot_acc(ax5, acc, eeg_start_time)

    # config
    plt.tight_layout()
    save_path = eeg_path.replace('eeg.eeg', 'data_preview.png')
    plt.savefig(save_path, dpi=300, bbox_inches='tight')


def plot_preview_mr2(eeg0, eeg1, acc, eeg_start_time, sf_eeg, eeg_path):
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(20, 3 * 4))
    # spectrogram
    t = plot_spectrogram(ax1, eeg0, eeg_start_time, sf_eeg)
    t = plot_spectrogram(ax2, eeg1, eeg_start_time, sf_eeg)
    # acc
    plot_acc(ax3, acc, eeg_start_time)

    # config
    plt.tight_layout()
    save_path = eeg_path.replace('eeg.eeg', 'data_preview.png')
    plt.savefig(save_path, dpi=300, bbox_inches='tight')


def plot_ecg(ax, ecg, start_time):

    sf = 100
    time_per_sample = 1 / sf / 3600
    t = np.arange(ecg.shape[0]) * time_per_sample
    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)
    ax.plot(timestamp_num, ecg, lw=1.5, color='r')

    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_ylabel("ECG [μV]", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})

    ax.set_ylim(-100, 100)

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim(timestamp[0], timestamp[-1])


def plot_emg(ax, emg, start_time):

    sf = 100

    time_per_sample = 1 / sf / 3600
    t = np.arange(emg.shape[0]) * time_per_sample
    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)
    ax.plot(timestamp_num, emg, lw=1.5, color='r')

    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_ylabel("EMG [μV]", fontdict={"fontsize": 18})
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})

    # 设置y轴的显示范围
    ax.set_ylim(-500, 500)

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim(timestamp[0], timestamp[-1])



"""
绘制慢波增强对比图
"""
import matplotlib.pyplot as plt
from qleegss.plot.plot_eeg_spectrogram import plot_spectrogram
import numpy as np
import matplotlib.dates as mdates
from datetime import timedelta, datetime
import scipy.signal as signal


def plot_sw(eeg, eeg_start_time, sf_eeg, eeg_path, idx):
    idx = idx[:(idx.size // 2) * 2]
    stim_idx = idx[0::2]
    sham_idx = idx[1::2]

    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(20, 5 + 2 + 10))
    # spectrogram
    t = plot_spectrogram(ax1, eeg, eeg_start_time, sf_eeg)

    # point
    plot_point(ax2, t, eeg_start_time, stim_points=stim_idx, sham_points=sham_idx)

    # sw
    plot_wave(ax3, eeg, sf_eeg, stim_idx, sham_idx)
    # save
    save_path = eeg_path.replace('eeg.eeg', 'sw_stim_sham_fig.png')
    fig.suptitle('STIM & SHAM', fontsize=25)
    plt.savefig(save_path, dpi=300, bbox_inches='tight')


def plot_point(ax, t, start_time, stim_points=None, sham_points=None):
    stim_points = stim_points / 100.0 / 3600
    sham_points = sham_points / 100.0 / 3600

    stim_points_timestamp = [start_time + timedelta(hours=hours) for hours in stim_points]
    stim_points_timestamp_num = mdates.date2num(stim_points_timestamp)

    sham_points_timestamp = [start_time + timedelta(hours=hours) for hours in sham_points]
    sham_points_timestamp_num = mdates.date2num(sham_points_timestamp)

    timestamp = [start_time + timedelta(hours=hours) for hours in t]
    timestamp_num = mdates.date2num(timestamp)

    ax.scatter(stim_points_timestamp_num, np.ones(len(stim_points)) * 15, c="red")
    ax.scatter(sham_points_timestamp_num, np.ones(len(stim_points)) * 5, c="gray")

    ax.set_ylim([0, 20])
    ax.set_yticks([5, 15])
    ax.set_yticklabels(["Sham", "Stim"])

    ax.tick_params(axis='y', labelsize=14)
    ax.tick_params(axis='x', labelsize=14)
    ax.set_xlabel("Time [hrs]", fontdict={"fontsize": 18})

    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M'))
    ax.set_xlim(timestamp_num[0], timestamp_num[-1])

    return ax


def plot_wave(ax, eeg, sf_eeg, stim_indicator, sham_indicator):
    eeg_filtered = eeg_filter(eeg, sf_eeg, 0.5, 3)

    seg_downlim = 1  # ERP down limitation
    seg_uplim = 4  # ERP up limitation

    EEGT = {'stim': np.array(
        [eeg_filtered[i - int(sf_eeg * seg_downlim):i + int(sf_eeg * seg_uplim)] for i in
         stim_indicator]),
        'sham': np.array(
        [eeg_filtered[i - int(sf_eeg * seg_downlim):i + int(sf_eeg * seg_uplim)] for i in
         sham_indicator])}

    stim_sem = np.std(EEGT['stim'], axis=0) / np.sqrt(len(EEGT['stim']) / 2)
    sham_sem = np.std(EEGT['sham'], axis=0) / np.sqrt(len(EEGT['sham']) / 2)

    t = np.arange(-100, 400) / sf_eeg
    eeg_stim_mean: object = np.mean(EEGT['stim'], axis=0)
    eeg_sham_mean = np.mean(EEGT['sham'], axis=0)

    ax.errorbar(t, eeg_stim_mean, yerr=stim_sem, label='STIM', alpha=0.1, color='r')

    ax.errorbar(t, eeg_sham_mean, yerr=sham_sem, label='SHAM', alpha=0.1, color='k')

    onset_idx = 100
    offset_idx = int(onset_idx + 1.075 * sf_eeg)
    ax.scatter(t[[onset_idx, offset_idx]], eeg_stim_mean[[onset_idx, offset_idx]], c='r',
               label='STIM Timings')
    ax.scatter(t[[onset_idx, offset_idx]], eeg_sham_mean[[onset_idx, offset_idx]], c='g',
               label='SHAM Timings')

    ax.set_xlim([min(t), max(t)])

    ax.set_xlabel('Time (s)', fontsize=25)
    ax.set_ylabel('Voltage (uV)', fontsize=25)
    ax.tick_params(labelsize=25)
    ax.legend(fontsize=20, loc='upper right')


def eeg_filter(eeg, sf_eeg, highpass, lowpass):

    wn_h = 2 * highpass / sf_eeg
    b_h, a_h = signal.butter(3, wn_h, 'highpass', analog=False)

    eeg = signal.lfilter(b_h, a_h, eeg)
    wn_l = 2 * lowpass / sf_eeg
    b_l, a_l = signal.butter(3, wn_l, 'lowpass', analog=False)

    eeg = signal.lfilter(b_l, a_l, eeg)

    return eeg


if __name__ == '__main__':
    eeg_ = np.random.rand(30000)
    idx_ = np.array([10000, 10100, 20000, 20100])
    plot_sw(eeg_, datetime.now(), 100, './eeg.eeg', idx_)
    plt.show()
